import logging
import math
import os
from typing import List

import torch
import torch.nn as nn

# Number of datapoints in storage on average
STORAGE_SUBSET_SIZE = 64


def get_gpu(func_id: int) -> int:
    """Based on the number of gpus, decide which one this function should use.

    Once the first function is run on the device, the environment variable for the gpu in that
    container will be set. If that is the case, return that id. If not, return the modulo of the function
    by the number of gpus to divide the computation equally"""

    # see if we already set the env variable
    gpu_id = os.getenv('GPU_ID', None)

    # if the env variable is set return the id
    if gpu_id is not None:
        logging.debug(f'ENV is set, using gpu {gpu_id} in function {func_id}')
        return int(gpu_id)

    # if not simply divide by the num of gpus and set the variable
    # for future usage
    gpu_count = torch.cuda.device_count()
    gpu_id = func_id % gpu_count
    os.environ['GPU_ID'] = str(gpu_id)
    logging.debug(f'Setting GPU_ID to {gpu_id} in function {func_id}')
    return gpu_id


def is_optimizable(layer: nn.Module) -> bool:
    """Should save layer returns just whether the layer is optimizable or not
    and thus if it should be sent to the parameter server"""
    t = str(type(layer))
    if 'conv' in t or 'linear' in t:
        return True
    return False


def split_minibatches(a: range, n: int) -> List[range]:
    """
    Based on the number of minibatches return the ones assigned to each
    function so that the count is approximately the same

    :arg a range with the list of minibatches
    :arg n number of functions to divide the minibatches across
    :return: list with all the ranges, indexed by the funcId
    """
    k, m = divmod(len(a), n)
    return [a[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n)]


def get_subset_period(K: int, batch_size: int, assigned_subsets: range) -> int:
    """
    Calculates the number of subsets that will be evaluated per iteration
    to fulfill the K-avg sync.

    Calculates the number of datapoints before sync, and from there calculates
    the number of subsets for those

    :param K: Number of forward passes to be done
    :param batch_size: size of the batch
    :param assigned_subsets: the data subsets assigned to this function

    :return: the number of subsets that must be loaded per iteration
    """

    # if K is -1, it means that we process all the data
    # assigned to us before syncing after a full epoch, so
    # return the length of the subsets
    if K == -1:
        return len(assigned_subsets)

    # calculate number of datapoints in K passes and divide to get the number of subsets
    return int(math.ceil((batch_size * K) / STORAGE_SUBSET_SIZE))



class KubeMLException(Exception):

    def __init__(self, message, status_code):
        Exception.__init__(self)
        self.message = message
        self.status_code = status_code

    def to_dict(self):
        return {
            "error": self.message,
            "code": self.status_code
        }


class MergeError(KubeMLException):
    def __init__(self, e: Exception = None):
        super(MergeError, self) \
            .__init__(f"Error merging model: {e}", 500)


class DataError(KubeMLException):
    def __init__(self):
        super(DataError, self) \
            .__init__("Data not present in request", 400)


class InvalidFormatError(KubeMLException):
    def __init__(self):
        super(InvalidFormatError, self) \
            .__init__("The data provided is not in an appropriate format", 400)


class StorageError(KubeMLException):
    def __init__(self, e: Exception):
        super(StorageError, self) \
            .__init__(f"Could not access storage service: {str(e)}", 500)


class DatasetNotFoundError(KubeMLException):
    def __init__(self):
        super(DatasetNotFoundError, self) \
            .__init__("Dataset not found in storage service", 404)


class InvalidArgsError(KubeMLException):
    def __init__(self, e: Exception):
        super(InvalidArgsError, self) \
            .__init__(f"Error parsing function arguments: {str(e)}", 500)



from abc import ABC
from collections import defaultdict
from typing import Dict, Tuple, Any, Union, Callable, Iterable, Sequence

import flask
import numpy as np
import pickle
import redisai as rai
import requests
from flask import request, jsonify, current_app
from redis.exceptions import RedisError
from torch.utils.data import DataLoader

from .dataset import _KubeArgs, KubeDataset
from .exceptions import *
from .util import *
import os

# Load from environment the values from th MONGO IP and PORT
try:
    REDIS_URL = os.environ['REDIS_URL']
    REDIS_PORT = os.environ['REDIS_PORT']
    logging.debug(f'Found configuration for redis {REDIS_URL}:{REDIS_PORT}')
except KeyError:
    logging.debug("Could not find redis configuration in env, using defaults")
    REDIS_URL = "redisai.kubeml"
    REDIS_PORT = 6379


class KubeModel(ABC):

    def __init__(self, network: nn.Module, dataset: KubeDataset, gpu=False):
        """Init the KubeModel, device can be either gpu or cpu"""

        # if device is set to gpu, get the correct gpu if
        # for the
        self._network = network
        self._dataset = dataset
        self.platform = 'gpu' if gpu else 'cpu'
        self.device = None
        self.args = None
        self.logger = None

        # training options, these will be updated when reading the parameters
        # in each iteration
        self.lr = None
        self.batch_size = None
        self.task = None
        self.optimizer = None
        self.epoch = None

        # initialize redis connection
        self._redis_client = rai.Client(host=REDIS_URL, port=REDIS_PORT)

    # allow to call the network from the kubemodel
    def __call__(self, *args, **kwargs):
        """
        Call allows users to invoke the underlying torch module by using the
        KubeModel. An example of this is when perrforming the forward pass:

        output = self(x)

        That the users can call in the train method

        :param args:
        :param kwargs:
        :return: the output of the network
        """
        return self._network(*args, **kwargs)

    # Functions that act like a proxy to the same
    # methods of the underlying pytorch module for convenience
    def parameters(self):
        """
        Parameters returns the parameter generator of the underlying network

        :return: The parameter generator of the torch module
        """
        return self._network.parameters()

    def apply(self, fn: Callable[[nn.Module], None]):
        """
        apply acts as a proxy to the underlying pytorch modules'
        apply method. It is mainly used to initialize weights of the network

        :param fn: function that initializes the layer weights
        :return: self
        """
        self._network.apply(fn)
        return self

    def _read_args(self):
        """Parse the args and update the parameters"""
        self.args = _KubeArgs.parse()
        self.lr = self.args.lr
        self.batch_size = self.args.batch_size
        self.task = self.args._task
        self.epoch = self.args.epoch

    def _config_optimizer(self):
        """
        Configures the optimizer specified by the user to be used in training. By default
        it resets the state of the optimizer like the momentum buffer in SGD with momentum
        :return:
        """
        optimizer = self.configure_optimizers()
        self.optimizer = optimizer
  
        # TODO here the state loaded should be the averaged one not the saved from earlier

        # self._load_optimizer_state()

    def _load_optimizer_state(self):
        """
        Checks for a previously saved state of the optimizer and loads it
        """
        if os.path.isfile('opt.pkl'):
            self.logger.debug('Loading optimizer state')
            with open('opt.pkl', 'rb') as f:
                self.logger.debug('Loading optimizer state, epoch: {self.epoch}')
                state = pickle.load(f)
                self.optimizer.load_state_dict(state)

    def _reset_optimizer_state(self):
        """
        Resets the optimizer state. Called at the start of each iteration
        since the momentum or other factors of the optimizer can hinder progress if
        carried over from another model
        """
        if self.optimizer is not None:
            self.optimizer.state = defaultdict(dict)

    def _save_optimizer_state(self):
        """
        Saves the optimizer state to be loaded again in
        the following epochs
        """
        
        self.logger.debug('saving optimizer state')

        with open('opt.pkl', 'wb') as f:
            self.logger.debug('saving optimizer state, epoch: {self.epoch}')
            pickle.dump(self.optimizer.state_dict(), f)
        print('saved state')

    def _save_file_test(self):
        """
        Saves the optimizer state to be loaded again in
        the following epochs
        """
        job_id = self.args._job_id
        func_id = self.args._func_id
        self.logger.debug('saving file to host')

        with open(f'/output/{job_id}:test:{func_id}.pkl', 'wb') as f:
            self.logger.debug(f'saving file, epoch: {self.epoch}')
            pickle.dump('test', f)
            self.logger.debug('file saved')
        print('file saved')   

    def _load_file_test(self):
        """
        Saves the optimizer state to be loaded again in
        the following epochs
        """
        job_id = self.args._job_id
        func_id = self.args._func_id
        self.logger.debug('loading file to host')

        with open(f'/output/{job_id}:test:{func_id}.pkl', 'wb') as f:
            self.logger.debug(f'loading file, epoch: {self.epoch}')
            state = pickle.load(f)
            self.logger.debug(f'file loaded, value: {state}')
            print('file loaded')    

    def _save_optimizer_redis(self):
        """
        Saves the optimizer to the redis
        """
        job_id = self.args._job_id
        func_id = self.args._func_id

        self.logger.debug("Saving optimizer to the redis")

        weight_key = f'{job_id}:optimizer:{func_id}' 
        pickled_val  = pickle.dumps(self.optimizer.state_dict())

        self.logger.debug(f"Saving optimizer value: {pickled_val}")

        self._redis_client.set(weight_key, pickled_val)


    def _load_optimizer_redis(self):
        """
        Checks for a previously saved state of the optimizer and loads it from redis
        """

        job_id = self.args._job_id
        func_id = self.args._func_id

        self.logger.debug("Load optimizer from the redis")

        weight_key = f'{job_id}:optimizer:{func_id}' 

        if  self._redis_client.exists(weight_key):
            epoch = self.args.epoch
            weight_key_print = f'{job_id}:print:{func_id}' 
            value = f'{job_id}:print:{func_id}/{epoch}' 
            self._redis_client.set(weight_key_print, value)

            pickled_val = self._redis_client.get(weight_key)
            opt_states = pickle.loads(pickled_val)  
            self.optimizer.load_state_dict(opt_states)



    def _save_redis_test(self):
        """
        Saves the optimizer to the redis
        """
        job_id = self.args._job_id
        func_id = self.args._func_id
        epoch = self.args.epoch

        self.logger.debug("Saving test case to the redis")

        weight_key = f'{job_id}:test:{func_id}' 
        value = f'{job_id}:test:{func_id}/{epoch}' 
        self._redis_client.set(weight_key, value)


    def _load_redis_test(self):
        """
        Checks for a previously saved state of the optimizer and loads it from redis
        """

        job_id = self.args._job_id
        func_id = self.args._func_id

        self.logger.debug("Load test case from the redis")

        weight_key = f'{job_id}:test:{func_id}' 
        pickled_val = self._redis_client.get(weight_key)
        self.logger.debug(f"loading value:  {pickled_val}")


    def _get_logger(self):
        """
        Sets the current logger within the network according to the flask context
        """
        self.logger = current_app.logger

    def start(self) -> Tuple[flask.Response, int]:
        """
        Start executes the function invoked by the user
        """
        # parse arguments and
        self._read_args()
        self._get_logger()

        if self.task == "init":
            layers = self.__initialize()
            return jsonify(layers), 200

        elif self.task == "train":
            loss = self.__train()
            return jsonify(loss=loss), 200

        elif self.task == "val":
            acc, loss, length = self.__validate()
            return jsonify(loss=loss, accuracy=acc, length=length), 200

        elif self.task == "infer":
            preds = self.__infer()
            return jsonify(predictions=preds), 200

        else:
            self._redis_client.close()
            raise KubeMLException(f"Task {self.task} not recognized", 400)

    def __initialize(self) -> List[str]:
        """
        Initializes the network

        :return: the names of the optimizable layers of the network, which will be saved in the reference model
        """
        try:
            self.init()
            # according to the comments, it is nothing to do with the training process.
            self.__save_model()

        except RedisError as re:
            raise StorageError(re)
        finally:
            self._redis_client.close()

        return [name for name in self._network.state_dict()]

    def _on_train_start(self):
        """
        Prepares the network for training
        :return:
        """
        self._set_device()
        self._network.train()
        self._config_optimizer()


    def _on_train_end(self):
        """
        Executed after the end of the training loop
        :return:
        """
        # self.__save_model()
        # self._save_optimizer_redis()
        pass
        # self._save_optimizer_state()

    def _on_iteration_start(self):
        """
        Called at the start of each iteration

        Load the model from the averaged storage, and
        reset optimizer state
        :return:
        """
        self.__load_model()
        self._load_optimizer_redis()
        self._load_redis_test()
        #self._load_file_test()

        #self._load_optimizer_state()

        #self._reset_optimizer_state()

    def _on_iteration_end(self):
        """
        Called at the end of each iteration
        :return:
        """
        self.__save_model()
        self._save_optimizer_redis()
        self._save_redis_test()
        #self._save_file_test()
        #self._save_optimizer_state()

    def _batch_to_device(self, batch: Union[torch.Tensor, Iterable[torch.Tensor]]):
        """
        Moves the batch fetched from the dataloader to the in use device. This allows to
        return an unbounded number of tensors from the dataset and passing those in the train step

        :return: The batch variables moved to the device
        """

        # this can be a list, tuple or tensor if it is just one
        elem_type = type(batch)

        # if it is a tensor, return to device directly
        if isinstance(batch, torch.Tensor):
            return batch.to(self.device)

        # if it is a tuple, return a tuple
        # with the elements to the device
        elif isinstance(batch, tuple):
            return elem_type((elem.to(self.device) for elem in batch))

        elif isinstance(batch, Sequence):
            return elem_type([elem.to(self.device) for elem in batch])

        else:
            return batch

    def __train(self) -> float:
        """
        Function called to train the network. Loads the reference model from the database,
        trains with the method provided by the user and saves the model after training to the database

        :return: The loss of the epoch, as returned by the user function
        """

        self._on_train_start()

        # Determine the batches that we need to train on and the first subset id
        assigned_subsets = split_minibatches(range(self._dataset.num_docs),
                                             self.args._N)[self.args._func_id]

        # calculate the number of subsets that we need to train on
        # per epoch
        subsets_per_iter = get_subset_period(self.args._K,
                                             self.args.batch_size,
                                             assigned_subsets)
        self.logger.debug(f"Subsets per iteration: {subsets_per_iter}")
        intervals = range(assigned_subsets.start, assigned_subsets.stop, subsets_per_iter)

        # the loss will be added cross intervals, each interval will have one loader, whose length
        # will determine the number of losses added.
        loss = 0
        num_iterations = 0
        for i in intervals:

            self.logger.debug(f"Starting iteration {i}")
            self._dataset._load_train_data(start=i, end=min(assigned_subsets.stop, i + subsets_per_iter))

            # create the loader that will be used
            loader = DataLoader(self._dataset, batch_size=self.batch_size)
            num_iterations += len(loader)

            # load the reference model, train and save
            try:
                self._on_iteration_start()

                for idx, batch in enumerate(loader):
                    # send the batch to the appropriate device
                    batch = self._batch_to_device(batch)
                    loss += self.train(batch, idx)
                    # self.logger.debug(f'loss is {loss}, iterations are {num_iterations}')

                self._on_iteration_end()
            except RedisError as re:
                raise StorageError(re)
            finally:
                self._redis_client.close()

            # send notification to the train job to refresh the model if not
            # the last interval
            if i != intervals[-1]:
                self.__send_finish_signal()

        self._on_train_end()

        return loss / num_iterations

    def _on_validation_start(self):
        """
        Executed before the validation
        :return:
        """
        self._set_device()
        self._network.eval()

    def __validate(self):
        """
        Validate sets the device to be used and sets the network in eval mode.
        Then it:
        - Loads the validation data
        - Creates a data loader
        - Feeds the validate function defined by the user with datapoints already sent to the correct device

        :return: A tuple containing the mean accuracy and loss on the val dataset and the number or datapoints
        """

        self._on_validation_start()

        # Determine the batches that we need to validate on and the first
        # subset id that we need to get each iteration
        assigned_subsets = split_minibatches(range(self._dataset.num_val_docs), self.args._N)[self.args._func_id]

        # load the validation data
        self._dataset._load_validation_data(start=assigned_subsets.start,
                                            end=assigned_subsets.stop)

        # create the loader that will be used
        loader = DataLoader(self._dataset, batch_size=self.batch_size)

        acc, loss = 0, 0
        try:
            self.__load_model()
            with torch.no_grad():
                for idx, batch in enumerate(loader):
                    batch = self._batch_to_device(batch)
                    _acc, _loss = self.validate(batch, idx)

                    # accumulate statistics
                    acc += _acc
                    loss += _loss
        except RedisError as re:
            raise StorageError(re)
        finally:
            self._redis_client.close()

        return acc / len(loader), loss / len(loader), len(self._dataset)

    def __infer(self) -> Union[torch.Tensor, np.ndarray, List[float]]:
        data_json = request.json
        if not data_json:
            self.logger.error("JSON not found in request")
            raise DataError

        preds = self.infer(self._network, data_json)

        if isinstance(preds, torch.Tensor):
            return preds.cpu().numpy().tolist()
        elif isinstance(preds, np.ndarray):
            return preds.tolist()
        elif isinstance(preds, list):
            return preds
        else:
            raise InvalidFormatError

    def _set_device(self):
        """Set device updates the used gpu or cpu based on the function id and the previously
        used devices"""

        # if the indicated platform is cpu,
        # set that as the torch device
        if self.platform == 'cpu':
            self.device = torch.device('cpu')

        # if it's gpu, get the appropriate one and put the model there
        else:
            gpu_id = get_gpu(self.args._func_id)
            self.device = torch.device(f'cuda:{gpu_id}')
            self._network = self._network.to(self.device)
            self.logger.debug(f'Set device to {self.device}')

    def __send_finish_signal(self):
        """Sends a request to the train job communicating that the iteration is over
        and the model is published in the database.

        The PS will not respond until all the functions have finished the step
        """

        # create the url for the job service
        url = f"http://job-{self.args._job_id}.kubeml/next/{self.args._func_id}"

        try:
            self.logger.debug(f"Sending request to {url}")
            resp = requests.post(url)
        except requests.ConnectionError as e:
            self.logger.error("error connecting to the train job")
            raise MergeError(e)

        if not resp.ok:
            self.logger.error(f"Received non OK message. Code:{resp.status_code}. Msg: {resp.content.decode()}")
            raise MergeError()

    def __load_model(self):
        """
        Loads the model from redis ai and applies it to the network
        """
        state_dict = self.__get_model_dict()
        self._network.load_state_dict(state_dict)
        self.logger.debug("Loaded state dict from redis")

    def __get_model_dict(self) -> Dict[str, torch.Tensor]:
        """
        Fetches the model weights from the tensor storage

        :return: The state dict of the reference model
        """
        job_id = self.args._job_id

        state = dict()
        for name in self._network.state_dict():
            # load each of the layers in the statedict
            weight_key = f'{job_id}:{name}'
            w = self._redis_client.tensorget(weight_key)
            # set the weight
            state[weight_key[len(job_id) + 1:]] = torch.from_numpy(w)

        #self.logger.debug(f'Layers are {state.keys()}')

        return state

    def __save_model(self):
        """
        Saves the model to the tensor storage
        """
        job_id = self.args._job_id
        task = self.args._task
        func_id = self.args._func_id

        self.logger.debug("Saving model to the database")
        with torch.no_grad():
            for name, layer in self._network.state_dict().items():
                # Save the weights
                weight_key = f'{job_id}:{name}' \
                    if task == 'init' \
                    else f'{job_id}:{name}/{func_id}'
                self._redis_client.tensorset(weight_key, layer.cpu().detach().numpy(), dtype='float32')

        self.logger.debug('Saved model to the database')

    def configure_optimizers(self) -> torch.optim.Optimizer:
        pass

    def init(self):
        pass

    def train(self, batch, batch_index: int) -> float:
        pass

    def validate(self, batch, batch_index: int) -> Tuple[float, float]:
        pass

    def infer(self, data: List[Any]) -> Union[torch.Tensor, np.ndarray, List[float]]:
        pass


import pickle
from abc import ABC, abstractmethod

import numpy as np
import torch.utils.data as data
from flask import request
from pymongo import MongoClient
from pymongo.errors import PyMongoError

from .exceptions import *
from .util import *

# Load from environment the values from th MONGO IP and PORT
try:
    MONGO_URL = os.environ['MONGO_IP']
    MONGO_PORT = os.environ['MONGO_PORT']
    logging.debug(f'Found configuration for storage {MONGO_URL}:{MONGO_PORT}')
except KeyError:
    logging.debug("Could not find mongo configuration in env, using defaults")
    MONGO_URL = "mongodb.kubeml"
    MONGO_PORT = 27017


class _KubeArgs:
    """
    Arguments used by the function to transmit the information needed to perform a
    training, validation or inference task. The arguments are:
    """

    def __init__(self, job_id: str,
                 N: int, K: int,
                 task: str,
                 func_id: int,
                 epoch: int,
                 lr: float = 0,
                 batch_size: int = 0,
                 ):
        """
        :arg job_id: id of the job\n
        :arg N: number of functions or parallelism
        :arg K: parameter for K-averaging, number of forward passes before sync
        :arg task: type of task (train, inference...)
        :arg func_id: id of the function
        :arg lr: learning rate
        :arg batch_size: size of the batch
        """

        self._job_id = job_id
        self._N = N
        self._K = K
        self._task = task
        self._func_id = func_id
        self.lr = lr
        self.batch_size = batch_size
        self.epoch = epoch

    @classmethod
    def parse(cls):
        """
        Parses the arguments from the request context
        :return: returns a KubeArgs object used by other methods
        """
        try:
            job_id = request.args.get("jobId")
            N = request.args.get("N", type=int)
            K = request.args.get("K", type=int)
            task = request.args.get("task")
            func_id = request.args.get("funcId", type=int)
            lr = request.args.get("lr", type=float)
            batch_size = request.args.get("batchSize", type=int)
            epoch = request.args.get("epoch", type=int)

        except ValueError as ve:
            logging.error(f"Error parsing request arguments: {ve}, args:{request.args}")
            raise InvalidArgsError(ve)

        args = cls(job_id, N, K, task, func_id, epoch, lr, batch_size)
        return args


class KubeDataset(data.Dataset, ABC):
    """
    KubeDataset is the main abstraction used by KubeML to load the data in a
    distributed manner from the storage in Kubernetes.
    The datasets created by the users need to override the class and call the super() in the init.
    After that, they must override the __getitem__ method like in a normal torch dataset
    The Kube Dataset will expose two properties, data and labels, which will be automatically
    loaded from the database
    """

    def __init__(self, dataset: str):
        """
        Init reads the data from the database given the dataset name

        :arg dataset Name of the dataset in the KubeML storage service
        """

        self.dataset = dataset
        self._mode = None
        self._client = MongoClient(MONGO_URL, MONGO_PORT)
        self._database = self._client[dataset]
        self._args = None

        # data and labels of the dataset
        self.data, self.labels = None, None

        # Check first if the dataset that the user gave as input
        # is available in the configured storage service
        try:
            dbs = set(self._client.list_database_names())
            if self.dataset not in dbs:
                logging.error(f"Dataset not in the storage service. "
                              f"Dataset = {dataset},"
                              f"Available = {dbs}")
                self._client.close()
                raise DatasetNotFoundError

        except PyMongoError as e:
            self._client.close()
            raise StorageError(e)

        # Set the range of minibatches that this function will train on and the ones
        # that will be used for validation
        self.num_docs = self._database["train"].count_documents({})
        self.num_val_docs = self._database["test"].count_documents({})
        logging.debug(f"Num docs: {self.num_docs}, Num val docs: {self.num_val_docs}")

    def _eval(self):
        """
        Sets the dataset in eval mode
        """
        self._mode = 'val'

    def _train(self):
        """
        Sets the dataset in train mode
        """
        self._mode = 'train'

    def is_training(self) -> bool:
        """
        Sees if the dataset is performing the training task. This is useful to
        do when you want to use different transformations for train and validation tasks.
        This way, you can call this function and depending of the result, use one or other transformation.

        :return: True if dataset is training, false otherwise
        """
        return self._mode == 'train'

    def _load_train_data(self, start: int, end: int):
        """
        For K averaging the data needs to be refreshed with the next K batches
        after every synchronization step, this is triggered by the KubeModel before
        starting another iteration

        :param start: first subset to be loaded
        :param end: last subset to be loaded
        """
        # load the minibatches given by the network
        minibatches = range(start, end)
        logging.debug(f"Loading minibatches {minibatches}")
        self.data, self.labels = self.__load_data(minibatches)

        # put the dataset in train mode
        self._train()

    def _load_validation_data(self, start: int, end: int):
        """
        Loads the validation data given the subsets assigned to this functions

        :param start: first subset to be loaded
        :param end: last subset to be loaded
        """
        minibatches = range(start, end)
        logging.debug(f"Loading minibatches {minibatches}")
        self.data, self.labels = self.__load_data(minibatches, validation=True)

        # put the dataset in validation mode
        self._eval()

    def _close(self):
        self._client.close()

    def __load_data(self, minibatches: range, validation=False):
        """
        Load the data needed to perform the train or validation tasks.

        Based on the minibatches, load the validation or train subsets that are
        within the limits of the given range

        :param minibatches: start and stop ids of the subsets that we must load
        :param validation: whether to load the validation data instead of the train data
        :return: the numpy arrays holding the features and labels of the data
        """

        try:
            # based on the validation flag load the data from a different collection
            collection = 'test' if validation else 'train'
            batches = self._database[collection].find({
                '_id': {
                    '$gte': minibatches.start,
                    '$lte': minibatches.stop - 1}
            })
        except PyMongoError as e:
            self._client.close()
            raise StorageError(e)

        data, labels = None, None
        for batch in batches:

            # load the data and the labels as numpy arrays
            # flatten the labels so instead of i.e (64,1) they're (64,)
            # thus we can call hstack on them
            d = pickle.loads(batch['data'])
            l = pickle.loads(batch['labels'])

            if data is None:
                data, labels = d, l.flatten()
            else:
                data = np.vstack([data, d])
                labels = np.hstack([labels, l.flatten()])

        return data, labels.flatten()

    @abstractmethod
    def __len__(self):
        pass



from setuptools import setup
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='serverlessdl',
    version='0.5.6',
    description='Python tools for training Neural Networks in a serverless setup',
    author='Ning Wang',
    author_email="nwang@futurewei.com",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=['serverlessdl'],
    install_requires=[
        'torch>=1.7',
        'redisai>=1.0.1',
        'pymongo>=3.11.1',
        'flask>=1.1.2'
    ],
    license="MIT",
)
